#!/usr/bin/env bash

declare    scriptDirectory=$(dirname "${0}") scriptName=$(basename "${0}")

declare    zeroPad="0000" IFS=$'\n' tiles=":" directionX directionY last threeByThree withoutWater
declare -a elevations offsets riversOverlay terrainLayer
declare -i i j x y number mapWidth=${1:-50} mapHeight=${2:-50} elevation offset increment howManyRivers

ChooseTile() {
	tile=" "
	case ${elevation} in
		0	)	tiles=" " ;;							# Shallow ocean only for trireme path around island
		1	)	tiles="ggggggggggggggggggggggggjjjjjjjjjjjjjjjjjjjjdpp" ;;	# coastal - favor grassland over desert (beach)
		2	)	tiles="gggggggggggggggggggsffhhjjjjjjjggggggggggffhhjj+" ;;	# inland lowlands
		3	)	tiles="ppppppppppggffhhjjppppppppppggffhhjj+" ;;		# high plains
		4	)	tiles="hhhhhhhhhhppffjjhhhhhhhhhhppffjj+" ;;			# hilly
		5	)	tiles="mmmmmhhhffjjmmmmmhhhffjj+" ;;				# mountainous
		6	)	tiles="mmat" ;;							# mountainous, very high possibly with snow caps
	esac
	tile="${tiles:$((${RANDOM}%${#tiles})):1}"
}

ChangeXY() {
	local number=$((${RANDOM}%10)) 
	if [ ${number} -lt 5 ] ; then
		[ "${last:0:1}" = "x" ] && last="y${last:1}"
	elif [ ${number} -gt 8 ] ; then
		[ "${last:0:1}" = "y" ] && last="x${last:1}"
	fi

	if [ "${last:0:1}" = "y" ] ; then
		# change x
		if [ "${directionX}" = "add" ] ; then
			((x+=1))
			last="x+"
		else
			((x-=1))
			last="x-"
		fi
	else
		# change y
		if [ "${directionY}" = "add" ] ; then
			((y+=1))
			last="y+"
		else
			((y-=1))
			last="y-"
		fi
	fi
}

AddRiverTile() {
	# Check current x,y on terrain and river overlay layers
	if [ "${terrainLayer[${y}]:${x}:1}" = " " ] || [ "${terrainLayer[${y}]:${x}:1}" = ":" ] || [ "${terrainLayer[${y}]:${x}:1}" = "+" ] ; then
		return
	elif [ "${riversOverlay[${y}]:${x}:1}" = "2" ] ; then
		return 1
	else
		# change current chr at x,y to "2"
		riversOverlay[${y}]="${riversOverlay[${y}]:0:${x}}2${riversOverlay[${y}]:$((${x}+1))}"
	fi
}

if [ ! -z "${1}" ] ; then
	while [ "${1}" != "" ]; do
		case "${1}" in
			-x | -X | -w | -W | 	--width		)	shift ; mapWidth="${1}" ;;
			-y | -Y | -h | -H | 	--height	)	shift ; mapHeight="${1}" ;;
			* )		echo "Usage: ${0} -x 100 -y 50" ; exit 1 ;;
		esac
	shift
	done
fi

# Check map width/height minimums
if [ ${mapWidth} -lt 16 -o ${mapHeight} -lt 16 ] ; then
	echo "FreeCiv minimum map width/height is 16"
	exit 1
fi

# Generate terrain layer with elevation at 0 (ocean)
terrainLayer[0]=$(z=$(printf "%*s" ${mapWidth} "") ; echo "${z// /0}")

for ((i=1;i<${mapHeight};i++)) ; do
	terrainLayer[${i}]=${terrainLayer[0]}
done

if [ ${mapHeight} -lt ${mapWidth} ] ; then
	increment=$(( (${mapHeight}/2)/6 ))
	else
	increment=$(( (${mapWidth}/2)/6 ))
fi

offsets=( 2 $((${increment}*2)) $((${increment}*3)) $((${increment}*4)) $((${increment}*5)) $((${increment}*6))  )
elevations=(1 2 3 4 5 6)

# Alter terrain layer with elevation at ${k}
for ((k=0;k<${#offsets[*]};k++)) ; do
	offset=${offsets[${k}]}
	elevation=${elevations[${k}]}
	for ((i=${offset};i<$((${mapHeight}-${offset}));i++)) ; do
		for ((j=${offset};j<$((${mapWidth}-${offset}));j++)) ; do
			# round the corners
			[ ${i} -eq ${offset} -a ${j} -eq ${offset} ] && continue
			[ ${i} -eq ${offset} -a ${j} -eq $(((${mapWidth}-${offset})-1)) ] && continue
			[ ${i} -eq $(((${mapHeight}-${offset})-1)) ] && [ ${j} -eq ${offset} ] && continue
			[ ${i} -eq $(((${mapHeight}-${offset})-1)) ] && [ ${j} -eq $(( (${mapWidth}-${offset})-1 )) ] && continue
			terrainLayer[${i}]="${terrainLayer[${i}]:0:$((${j}))}${elevation}${terrainLayer[${i}]:$((${j}+1))}"
		done
	done
done



# Rough-up the coastline

# Horizontal coastlines
for i in ${offsets[0]} $(( ${mapHeight} - ${offsets[0]} -1 )) ; do
	for ((j=${offsets[0]};j<=$((${mapWidth}-${offsets[0]}));j++ )) ; do
		# 20% chance of making the coastline tile ocean
		if [ $((${RANDOM}%5)) -eq 0 -a "${terrainLayer[${i}]:${j}:1}" = "1" ] ; then
			terrainLayer[${i}]="${terrainLayer[${i}]:0:${j}}0${terrainLayer[${i}]:$((${j}+1))}"
			# 33% chance of changing the newly coastal land tile to ocean too
			if [ $((${RANDOM}%3)) -eq 0 ] ; then
				if [ ${i} -gt $((${mapHeight}/2)) -a "${terrainLayer[$((${i}-1))]:${j}:1}" = "1" ] ; then 
					terrainLayer[$((${i}-1))]="${terrainLayer[$((${i}-1))]:0:${j}}0${terrainLayer[$((${i}-1))]:$((${j}+1))}"
				elif  [ ${i} -lt $((${mapHeight}/2)) -a "${terrainLayer[$((${i}+1))]:${j}:1}" = "1" ] ; then
					terrainLayer[$((${i}+1))]="${terrainLayer[$((${i}+1))]:0:${j}}0${terrainLayer[$((${i}+1))]:$((${j}+1))}"
				fi
			fi
		fi
	done
done

# Vertical coastlines
for j in ${offsets[0]} $(( ${mapWidth} - ${offsets[0]} - 1 )) ; do
	for ((i=${offsets[0]};i<=$((${mapHeight}-${offsets[0]}));i++ )) ; do
		# 20% chance of making the coastline tile ocean
		if [ $((${RANDOM}%5)) -eq 0 -a "${terrainLayer[${i}]:${j}:1}" = "1" ] ; then
			terrainLayer[${i}]="${terrainLayer[${i}]:0:${j}}0${terrainLayer[${i}]:$((${j}+1))}"
			# 33% chance of changing the newly coastal land tile to ocean too
			if [ $((${RANDOM}%3)) -eq 0 ] ; then
				if [ ${j} -gt $((${mapWidth}/2)) -a "${terrainLayer[${i}]:$((${j}-1)):1}" = "1" ] ; then 
					terrainLayer[${i}]="${terrainLayer[${i}]:0:$((${j}-1))}0${terrainLayer[${i}]:${j}}"
				elif  [ ${i} -lt $((${mapWidth}/2)) -a "${terrainLayer[${i}]:$((${j}+1)):1}" = "1" ] ; then
					terrainLayer[${i}]="${terrainLayer[${i}]:0:$((${j}+1))}0${terrainLayer[${i}]:$((${j}+2))}"
				fi
			fi
		fi
	done
done

# Randomize terrain
for ((i=0;i<${mapHeight};i++)) ; do
	for ((j=0;j<${mapWidth};j++)) ; do
		elevation=${terrainLayer[${i}]:${j}:1}
		ChooseTile
		terrainLayer[${i}]="${terrainLayer[${i}]:0:$((${j}))}${tile}${terrainLayer[${i}]:$((${j}+1))}"
	done
done

# Remove 1x1 islands
for ((i=0;i<${mapHeight};i++)) ; do
	for ((j=0;j<$((${mapWidth}-1));j++)) ; do
		threeByThree="${terrainLayer[$((${i}-1))]:$((${j}-1)):3}${terrainLayer[${i}]:$((${j}-1)):3}${terrainLayer[$((${i}+1))]:$((${j}-1)):3}"
		withoutWater="${threeByThree//[ :+]}"
		[ ${#withoutWater} -eq 1 ] && terrainLayer[${i}]="${terrainLayer[${i}]:0:${j}}${terrainLayer[${i}]:$((${j}-1)):1}${terrainLayer[${i}]:$((${j}+1))}"
	done
done

# output terrain map
for ((i=0;i<${mapHeight};i++)) ; do
	echo "t${zeroPad:${#i}}${i}=\"${terrainLayer[${i}]}\""
done
echo

# Generate river overlay map layer
riversOverlay[0]=$(z=$(printf "%*s" ${mapWidth} "") ; echo "${z// /0}")
for ((i=1;i<${mapHeight};i++)) ; do
	riversOverlay[${i}]=${riversOverlay[0]}
done

howManyRivers=$(( ( ${RANDOM} % 3 ) + 2 )) # 2 to 4 rivers although if 2 and the 2nd river overlaps the first then only 1 river occurs
for ((i=0;i<${howManyRivers};i++)) ; do
	x=$(( (${mapWidth}/2) + ( (${RANDOM}%6) -3 ) ))
	y=$(( (${mapHeight}/2) + ( (${RANDOM}%6) -3 ) ))
	[ ${x} -gt $((${mapWidth}/2)) ] && directionX="add" || directionX="subtract"
	[ ${y} -gt $((${mapHeight}/2)) ] && directionY="add" || directionY="subtract"
	last="y+"
	while : ; do # until current river is done
		[ ${x} -lt 0 ] || [ ${x} -gt $((${mapWidth}-1)) ] || [ ${y} -lt 0 ] || [ ${y} -gt $((${mapHeight}-1)) ] && break
		AddRiverTile
		# If the current river ran into existing river, stop
		[ ${?} -eq 1 ] && break 2
		ChangeXY
	done
done

# output river overlay map
for ((i=0;i<${#riversOverlay[*]};i++)) ; do
	echo "e03_${zeroPad:${#i}}${i}=\"${riversOverlay[${i}]}\""
done

echo
echo "Width/xsize=${#terrainLayer[0]}  Height/ysize=${#terrainLayer[*]}"

exit 0
